# JUBE Benchmarking Environment
# Copyright (C) 2008-2020
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tabletype definition"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

from jube2.result_types.keyvaluesresult import KeyValuesResult
from jube2.result import Result
import xml.etree.ElementTree as ET
import jube2.log
import jube2.util.output

LOGGER = jube2.log.get_logger(__name__)


class Table(KeyValuesResult):

    """A ascii based result table"""

    class TableData(KeyValuesResult.KeyValuesData):

        """Table data"""

        def __init__(self, name_or_other, style, separator, transpose):
            if type(name_or_other) is KeyValuesResult.KeyValuesData:
                self._name = name_or_other.name
                self._keys = name_or_other.keys
                self._data = name_or_other.data
                self._benchmark_ids = name_or_other.benchmark_ids
            else:
                KeyValuesResult.KeyValuesData.__init__(self, name_or_other)
            self._style = style
            self._separator = separator
            # Ignore separator if pretty style is used
            if self._style == "pretty":
                self._separator = None
            elif self._separator is None:
                self._separator = jube2.conf.DEFAULT_SEPARATOR
            self._transpose = transpose

        @property
        def _columns(self):
            """Get columns"""
            return self._keys

        @property
        def style(self):
            """Get style"""
            return self._style

        @style.setter
        def style(self, style):
            """Set style"""
            self._style = style

        @property
        def separator(self):
            """Get separator"""
            return self._separator

        @separator.setter
        def separator(self, separator):
            """Set separator"""
            self._separator = separator

        def __str__(self):
            colw = list()
            for column in self._columns:
                if type(column) is Table.Column:
                    if column.colw is None:
                        colw.append(0)
                    else:
                        colw.append(column.colw)
                else:
                    colw.append(0)
            data = list()
            data.append([column.resulting_name for column in self._columns])
            data += self._data
            if self._style == "pretty":
                output = "{0}:\n".format(self.name)
            else:
                output = ""

            output += jube2.util.output.text_table(
                data, use_header_line=True, auto_linebreak=False, colw=colw,
                indent=0, style=self._style,
                separator=self._separator, transpose=self._transpose)

            return output

        def create_result(self, show=True, filename=None, **kwargs):
            """Create result output"""
            # If there are multiple benchmarks, add benchmark id information
            if len(set(self._benchmark_ids)) > 1:
                self.add_id_information(reverse=kwargs.get("reverse", False))

            result_str = str(self)

            # Print result to screen
            if show:
                LOGGER.info(result_str)
                LOGGER.info("\n")
            else:
                LOGGER.debug(result_str)
                LOGGER.debug("\n")

            # Print result to file
            if filename is not None:
                file_handle = open(filename, "w")
                file_handle.write(result_str)
                file_handle.close()

    class Column(KeyValuesResult.DataKey):

        """Class represents one table column"""

        def __init__(self, name, title=None, colw=None, format_string=None,
                     unit=None):
            KeyValuesResult.DataKey.__init__(self, name, title, format_string,
                                             unit)
            self._colw = colw

        @property
        def colw(self):
            """Column width"""
            return self._colw

        def etree_repr(self):
            """Return etree object representation"""
            column_etree = KeyValuesResult.DataKey.etree_repr(self)
            column_etree.tag = "column"
            if self._colw is not None:
                column_etree.attrib["colw"] = str(self._colw)
            return column_etree

    def __init__(self, name, style="csv",
                 separator=jube2.conf.DEFAULT_SEPARATOR,
                 sort_names=None,
                 transpose=False,
                 res_filter=None):
        KeyValuesResult.__init__(self, name, sort_names, res_filter)
        self._style = style
        self._separator = separator
        self._transpose = transpose

    def add_column(self, name, colw=None, format_string=None, title=None):
        """Add an additional column to the dataset"""
        self._keys.append(Table.Column(name, title, colw, format_string))

    def add_key(self, name, format_string=None, title=None, unit=None):
        """Add an additional key to the dataset"""
        self._keys.append(Table.Column(name, title, None, format_string))

    def create_result_data(self, style):
        """Create result data"""
        result_data = KeyValuesResult.create_result_data(self)
        return Table.TableData(result_data,
                               style if style is not None else self._style,
                               self._separator, self._transpose)

    def etree_repr(self):
        """Return etree object representation"""
        result_etree = Result.etree_repr(self)
        table_etree = ET.SubElement(result_etree, "table")
        table_etree.attrib["name"] = self._name
        table_etree.attrib["style"] = self._style
        if self._separator is not None:
            table_etree.attrib["separator"] = self._separator
        if self._res_filter is not None:
            table_etree.attrib["filter"] = self._res_filter
        table_etree.attrib["transpose"] = str(self._transpose)
        if len(self._sort_names) > 0:
            table_etree.attrib["sort"] = \
                jube2.conf.DEFAULT_SEPARATOR.join(self._sort_names)
        for column in self._keys:
            table_etree.append(column.etree_repr())
        return result_etree



# JUBE Benchmarking Environment
# Copyright (C) 2008-2020
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""CLI program"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import jube2.jubeio
import jube2.util.util
import jube2.util.output
import jube2.conf
import jube2.info
import jube2.help
import jube2.log
import jube2.completion

import sys
import os
import re
import shutil
from distutils.version import StrictVersion

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen

try:
    import argparse
except ImportError:
    print("argparse module not available; either install it "
          "(https://pypi.python.org/pypi/argparse), or "
          "switch to a Python version that includes it.")
    sys.exit(1)

LOGGER = jube2.log.get_logger(__name__)


def continue_benchmarks(args):
    """Continue benchmarks"""
    found_benchmarks = search_for_benchmarks(args)
    jube2.conf.HIDE_ANIMATIONS = args.hide_animation
    for benchmark_folder in found_benchmarks:
        _continue_benchmark(benchmark_folder, args)


def status(args):
    """Show benchmark status"""
    found_benchmarks = search_for_benchmarks(args)
    for benchmark_folder in found_benchmarks:
        benchmark = _load_existing_benchmark(args, benchmark_folder,
                                             load_analyse=False)
        if benchmark is None:
            return
        jube2.info.print_benchmark_status(benchmark)


def benchmarks_results(args):
    """Show benchmark results"""
    found_benchmarks = search_for_benchmarks(args)
    result_list = list()

    # Start with the newest benchmark to set the newest result configuration
    found_benchmarks.reverse()
    cnt = 0
    for benchmark_folder in found_benchmarks:
        if (args.num is None) or (cnt < args.num):
            result_list = _benchmark_result(benchmark_folder=benchmark_folder,
                                            args=args,
                                            result_list=result_list)
            cnt += 1
    for result_data in result_list:
        result_data.create_result(reverse=args.reverse)


def analyse_benchmarks(args):
    """Analyse benchmarks"""
    found_benchmarks = search_for_benchmarks(args)
    for benchmark_folder in found_benchmarks:
        _analyse_benchmark(benchmark_folder, args)


def remove_benchmarks(args):
    """Remove benchmarks or workpackages"""
    if(args.workpackage is not None):
        # If a workpackage id is provided by the user, only specific
        # workpackages will be removed
        found_workpackages = search_for_workpackage(args)
        for workpackage in found_workpackages:
            _remove_workpackage(workpackage, args)
    else:
        # Delete complete benchmarks
        found_benchmarks = search_for_benchmarks(args)
        for benchmark_folder in found_benchmarks:
            _remove_benchmark(benchmark_folder, args)


def command_help(args):
    """Show command help"""
    subparser = _get_args_parser()[1]
    if args.command is None:
        subparser["help"].print_help()
    elif args.command.lower() == "all":
        for key in sorted(jube2.help.HELP.keys()):
            print("{0}:".format(key))
            print(jube2.help.HELP[key])
    else:
        if args.command in jube2.help.HELP:
            if args.command in subparser:
                subparser[args.command].print_help()
            else:
                print(jube2.help.HELP[args.command])
        else:
            print("no help found for {0}".format(args.command))
            subparser["help"].print_help()


def info(args):
    """Benchmark information"""
    if args.id is None:
        jube2.info.print_benchmarks_info(args.dir)
    else:
        found_benchmarks = search_for_benchmarks(args)
        for benchmark_folder in found_benchmarks:
            benchmark = \
                _load_existing_benchmark(args, benchmark_folder,
                                         load_analyse=False)
            if benchmark is None:
                continue
            if args.step is None:
                jube2.info.print_benchmark_info(benchmark)
            else:
                if args.step:
                    steps = args.step
                else:
                    steps = benchmark.steps.keys()
                # Set default csv_parametrization value to allow empty -c
                # option
                if args.csv_parametrization is None:
                    args.csv_parametrization = ","
                for step_name in steps:
                    jube2.info.print_step_info(
                        benchmark, step_name,
                        parametrization_only=args.parametrization,
                        parametrization_only_csv=args.csv_parametrization)


def update_check(args):
    """Check if a newer JUBE version is available."""
    try:
        website = urlopen(jube2.conf.UPDATE_VERSION_URL)
        version = website.read().decode().strip()
        if StrictVersion(jube2.conf.JUBE_VERSION) >= StrictVersion(version):
            LOGGER.info("Newest JUBE version {0} is already "
                        "installed.".format(jube2.conf.JUBE_VERSION))
        else:
            LOGGER.info(("Newer JUBE version {0} is available. "
                         "Currently installed version is {1}.\n"
                         "New version can be "
                         "downloaded here: {2}").format(
                version, jube2.conf.JUBE_VERSION,
                jube2.conf.UPDATE_URL))
    except IOError as ioe:
        raise IOError("Cannot connect to {0}: {1}".format(
            jube2.conf.UPDATE_VERSION_URL, str(ioe)))
    except ValueError as verr:
        raise ValueError("Cannot read version string from {0}: {1}".format(
            jube2.conf.UPDATE_VERSION_URL, str(verr)))


def show_log(args):
    """Show logs for benchmarks"""
    found_benchmarks = search_for_benchmarks(args)
    for benchmark_folder in found_benchmarks:
        show_log_single(args, benchmark_folder)


def show_log_single(args, benchmark_folder):
    """Show logs for a single benchmark"""
    # Find available logs
    available_logs = jube2.log.search_for_logs(benchmark_folder)

    # Use all available logs if none is selected ...
    if not args.command:
        matching = available_logs
        not_matching = list()
    # ... otherwise find intersection between available and
    # selected
    else:
        matching, not_matching = jube2.log.matching_logs(
            args.command, available_logs)

    # Output the log file
    for log in matching:
        jube2.log.log_print("BenchmarkID: {0} | Log: {1}".format(
            int(os.path.basename(benchmark_folder)), log))
        jube2.log.safe_output_logfile(log)

    # Inform user if any selected log was not found
    if not_matching:
        jube2.log.log_print("Could not find logs: {0}".format(
            ",".join(not_matching)))


def complete(args):
    """Handle shell completion"""
    jube2.completion.complete_function_bash(args)


def _load_existing_benchmark(args, benchmark_folder, restore_workpackages=True,
                             load_analyse=True):
    """Load an existing benchmark, given by directory benchmark_folder."""

    jube2.log.change_logfile_name(os.path.join(
        benchmark_folder, jube2.conf.LOGFILE_PARSE_NAME))

    # Add log information
    LOGGER.debug("Command: {0} {1}".format(
        os.path.basename(sys.argv[0]), " ".join(sys.argv[1:])))
    LOGGER.debug("Version: {0}".format(jube2.conf.JUBE_VERSION))

    # Read existing benchmark configuration
    try:
        parser = jube2.jubeio.Parser(os.path.join(
            benchmark_folder, jube2.conf.CONFIGURATION_FILENAME),
            force=args.force, strict=args.strict)
        benchmarks = parser.benchmarks_from_xml()[0]
    except IOError as exeption:
        LOGGER.warning(str(exeption))
        return None

    # benchmarks can be None if version conflict was blocked
    if benchmarks is not None:
        # Only one single benchmark exist inside benchmarks
        benchmark = list(benchmarks.values())[0]
    else:
        return None

    # Restore old benchmark id
    benchmark.id = int(os.path.basename(benchmark_folder))

    if restore_workpackages:
        # Read existing workpackage information
        try:
            parser = jube2.jubeio.Parser(os.path.join(
                benchmark_folder, jube2.conf.WORKPACKAGES_FILENAME),
                force=args.force, strict=args.strict)
            workpackages, work_stat = parser.workpackages_from_xml(benchmark)
        except IOError as exeption:
            LOGGER.warning(str(exeption))
            return None
        benchmark.set_workpackage_information(workpackages, work_stat)

    if load_analyse and os.path.isfile(os.path.join(
            benchmark_folder, jube2.conf.ANALYSE_FILENAME)):
        # Read existing analyse data
        parser = jube2.jubeio.Parser(os.path.join(
            benchmark_folder, jube2.conf.ANALYSE_FILENAME),
            force=args.force, strict=args.strict)
        analyse_result = parser.analyse_result_from_xml()
        if analyse_result is not None:
            for analyser in benchmark.analyser.values():
                if analyser.name in analyse_result:
                    analyser.analyse_result = analyse_result[analyser.name]

    jube2.log.only_console_log()

    return benchmark


def manipulate_comments(args):
    """Manipulate benchmark comment"""
    found_benchmarks = search_for_benchmarks(args)
    for benchmark_folder in found_benchmarks:
        _manipulate_comment(benchmark_folder, args)


def search_for_benchmarks(args):
    """Search for existing benchmarks"""
    found_benchmarks = list()
    if not os.path.isdir(args.dir):
        raise OSError("Not a directory: \"{0}\"".format(args.dir))
    all_benchmarks = [
        os.path.join(args.dir, directory)
        for directory in os.listdir(args.dir)
        if os.path.isdir(os.path.join(args.dir, directory))]
    all_benchmarks.sort()
    if (args.id is not None) and ("all" not in args.id):
        for benchmark_id in args.id:
            if benchmark_id == "last":
                benchmark_id = jube2.util.util.get_current_id(args.dir)
            # Search for existing benchmark
            benchmark_id = int(benchmark_id)
            if benchmark_id < 0:
                benchmark_id = int(
                    os.path.basename(all_benchmarks[benchmark_id]))
            benchmark_folder = jube2.util.util.id_dir(args.dir, benchmark_id)
            if not os.path.isdir(benchmark_folder):
                raise OSError("Benchmark directory not found: \"{0}\""
                              .format(benchmark_folder))
            if not os.path.isfile(os.path.join(
                    benchmark_folder, jube2.conf.CONFIGURATION_FILENAME)):
                LOGGER.warning(("Configuration file \"{0}\" not found in " +
                                "\"{1}\" or directory not readable.")
                               .format(jube2.conf.CONFIGURATION_FILENAME,
                                       benchmark_folder))
            if benchmark_folder not in found_benchmarks:
                found_benchmarks.append(benchmark_folder)
    else:
        if (args.id is not None) and ("all" in args.id):
            # Add all available benchmark folder
            found_benchmarks = all_benchmarks
        else:
            # Get highest benchmark id and build benchmark_folder
            benchmark_id = jube2.util.util.get_current_id(args.dir)
            benchmark_folder = jube2.util.util.id_dir(args.dir, benchmark_id)
            if os.path.isdir(benchmark_folder):
                found_benchmarks.append(benchmark_folder)
            else:
                raise OSError("No benchmark directory found in \"{0}\""
                              .format(args.dir))

    found_benchmarks = \
        [benchmark_folder for benchmark_folder in found_benchmarks if
         os.path.isfile(os.path.join(benchmark_folder,
                                     jube2.conf.CONFIGURATION_FILENAME))]

    found_benchmarks.sort()
    return found_benchmarks


def search_for_workpackage(args):
    """Search for existing workpackages"""
    found_benchmarks = search_for_benchmarks(args)
    found_workpackages = list()
    for benchmark_folder in found_benchmarks:
        benchmark = \
            _load_existing_benchmark(args, benchmark_folder,
                                     load_analyse=False)
        if benchmark is not None:
            for wp_id in args.workpackage:
                if benchmark.workpackage_by_id(int(wp_id)) is None:
                    raise RuntimeError(("No workpackage \"{0}\" found " +
                                        "in benchmark \"{1}\".")
                                       .format(wp_id, benchmark.id))
                else:
                    found_workpackages.append(
                        benchmark.workpackage_by_id(int(wp_id)))
    return found_workpackages


def run_new_benchmark(args):
    """Start a new benchmark run"""

    jube2.conf.HIDE_ANIMATIONS = args.hide_animation
    jube2.conf.EXIT_ON_ERROR = args.error

    id_cnt = 0

    # Extract tags
    tags = args.tag
    if tags is not None:
        tags = set(tags)

    for path in args.files:
        # Setup Logging
        jube2.log.change_logfile_name(
            filename=os.path.join(os.path.dirname(path),
                                  jube2.conf.DEFAULT_LOGFILE_NAME))

        # Add log information
        LOGGER.debug("Command: {0} {1}".format(
            os.path.basename(sys.argv[0]), " ".join(sys.argv[1:])))
        LOGGER.debug("Version: {0}".format(jube2.conf.JUBE_VERSION))

        # Read new benchmarks
        if args.include_path is not None:
            include_pathes = [include_path for include_path in
                              args.include_path if include_path != ""]
        else:
            include_pathes = None
        parser = jube2.jubeio.Parser(path, tags, include_pathes,
                                     args.force, args.strict)
        benchmarks, only_bench, not_bench = parser.benchmarks_from_xml()

        # Add new comment
        if args.comment is not None:
            for benchmark in benchmarks.values():
                benchmark.comment = re.sub(r"\s+", " ", args.comment)

        # CLI input overwrite fileinput
        if args.only_bench:
            only_bench = args.only_bench
        if args.not_bench:
            not_bench = args.not_bench

        # No specific -> do all
        if len(only_bench) == 0 and benchmarks is not None:
            only_bench = list(benchmarks)

        for bench_name in only_bench:
            if bench_name in not_bench:
                continue
            bench = benchmarks[bench_name]
            # Set user defined id
            if (args.id is not None) and (len(args.id) > id_cnt):
                if args.id[id_cnt] < 0:
                    LOGGER.warning("Negative ids are not allowed. Skipping id "
                                   "'{}'.".format(args.id[id_cnt]))
                    id_cnt += 1
                    continue
                bench.id = args.id[id_cnt]
                id_cnt += 1
            # Change runtime outpath if specified
            if args.outpath is not None:
                bench.outpath = args.outpath
            # Start benchmark run
            bench.new_run()
            # Run analyse
            if args.analyse or args.result:
                jube2.log.change_logfile_name(os.path.join(
                    bench.bench_dir, jube2.conf.LOGFILE_ANALYSE_NAME))
                bench.analyse()

            # Create result data
            if args.result:
                jube2.log.change_logfile_name(os.path.join(
                    bench.bench_dir, jube2.conf.LOGFILE_RESULT_NAME))
                bench.create_result(show=True)

            # Clean up when using debug mode
            if jube2.conf.DEBUG_MODE:
                bench.delete_bench_dir()

        # Reset logging
        jube2.log.only_console_log()


def _continue_benchmark(benchmark_folder, args):
    """Continue existing benchmark"""

    jube2.conf.EXIT_ON_ERROR = args.error

    benchmark = _load_existing_benchmark(args, benchmark_folder)

    if benchmark is None:
        return

    # Change logfile
    jube2.log.change_logfile_name(os.path.join(
        benchmark_folder, jube2.conf.LOGFILE_CONTINUE_NAME))

    # Run existing benchmark
    benchmark.run()

    # Run analyse
    if args.analyse or args.result:
        jube2.log.change_logfile_name(os.path.join(
            benchmark_folder, jube2.conf.LOGFILE_ANALYSE_NAME))
        benchmark.analyse()

    # Create result data
    if args.result:
        jube2.log.change_logfile_name(os.path.join(
            benchmark_folder, jube2.conf.LOGFILE_RESULT_NAME))
        benchmark.create_result(show=True)

    # Clean up when using debug mode
    if jube2.conf.DEBUG_MODE:
        benchmark.reset_all_workpackages()

    # Reset logging
    jube2.log.only_console_log()


def _analyse_benchmark(benchmark_folder, args):
    """Analyse existing benchmark"""
    benchmark = _load_existing_benchmark(args, benchmark_folder,
                                         load_analyse=False)
    if benchmark is None:
        return

    # Update benchmark data
    _update_analyse_and_result(args, benchmark)

    # Change logfile
    jube2.log.change_logfile_name(os.path.join(
        benchmark_folder, jube2.conf.LOGFILE_ANALYSE_NAME))

    LOGGER.info(jube2.util.output.text_boxed(
        ("Analyse benchmark \"{0}\" id: {1}").format(benchmark.name,
                                                     benchmark.id)))
    benchmark.analyse()
    if os.path.isfile(
            os.path.join(benchmark_folder, jube2.conf.ANALYSE_FILENAME)):
        LOGGER.info(">>> Analyse data storage: {0}".format(os.path.join(
            benchmark_folder, jube2.conf.ANALYSE_FILENAME)))
    else:
        LOGGER.info(">>> Analyse data storage \"{0}\" not created!".format(
            os.path.join(benchmark_folder, jube2.conf.ANALYSE_FILENAME)))
    LOGGER.info(jube2.util.output.text_line())

    # Reset logging
    jube2.log.only_console_log()


def _benchmark_result(benchmark_folder, args, result_list=None):
    """Show benchmark result"""
    benchmark = _load_existing_benchmark(args, benchmark_folder)
    if result_list is None:
        result_list = list()

    if benchmark is None:
        return result_list

    if (args.update is None) and (args.tag is not None) and \
            (len(benchmark.tags & set(args.tag)) == 0):
        return result_list

    # Update benchmark data
    _update_analyse_and_result(args, benchmark)

    # Run benchmark analyse
    if args.analyse:
        jube2.log.change_logfile_name(os.path.join(
            benchmark_folder, jube2.conf.LOGFILE_ANALYSE_NAME))
        benchmark.analyse(show_info=False)

        # Change logfile
    jube2.log.change_logfile_name(os.path.join(
        benchmark_folder, jube2.conf.LOGFILE_RESULT_NAME))

    # Create benchmark results
    result_list = benchmark.create_result(only=args.only,
                                          data_list=result_list,
                                          style=args.style)

    # Reset logging
    jube2.log.only_console_log()

    return result_list


def _update_analyse_and_result(args, benchmark):
    """Update analyse and result data in given benchmark by using the
    given update file"""
    if args.update is not None:
        dirname = os.path.dirname(args.update)
        # Extract tags
        benchmark.add_tags(args.tag)
        tags = benchmark.tags

        # Read new benchmarks
        if args.include_path is not None:
            include_pathes = [include_path for include_path in
                              args.include_path if include_path != ""]
        else:
            include_pathes = None
        parser = jube2.jubeio.Parser(args.update, tags, include_pathes,
                                     args.force, args.strict)
        benchmarks = parser.benchmarks_from_xml()[0]

        # Update benchmark
        for bench in benchmarks.values():
            if bench.name == benchmark.name:
                benchmark.update_analyse_and_result(bench.patternsets,
                                                    bench.analyser,
                                                    bench.results,
                                                    bench.results_order,
                                                    dirname)
                break
        else:
            LOGGER.debug(("No benchmark data for benchmark {0} was found " +
                          "while running update.").format(benchmark.name))


def _remove_benchmark(benchmark_folder, args):
    """Remove existing benchmark"""
    remove = True
    if not args.force:
        try:
            inp = raw_input("Really remove \"{0}\" (y/n):"
                            .format(benchmark_folder))
        except NameError:
            inp = input("Really remove \"{0}\" (y/n):"
                        .format(benchmark_folder))
        remove = inp.startswith("y")
    if remove:
        # Delete benchmark folder
        shutil.rmtree(benchmark_folder, ignore_errors=True)


def _remove_workpackage(workpackage, args):
    """Remove existing workpackages"""
    remove = True
    # Ignore deleted/unstarted workpackages
    if workpackage.started:
        if not args.force:
            try:
                inp = raw_input(("Really remove \"{0}\" and its dependent " +
                                 "workpackages (y/n):")
                                .format(workpackage.workpackage_dir))
            except NameError:
                inp = input(("Really remove \"{0}\" and its dependent " +
                             "workpackages (y/n):")
                            .format(workpackage.workpackage_dir))
            remove = inp.startswith("y")
        if remove:
            workpackage.remove()
            workpackage.benchmark.write_workpackage_information(
                os.path.join(workpackage.benchmark.bench_dir,
                             jube2.conf.WORKPACKAGES_FILENAME))


def _manipulate_comment(benchmark_folder, args):
    """Change or append the comment in given benchmark."""
    benchmark = _load_existing_benchmark(args,
                                         benchmark_folder=benchmark_folder,
                                         restore_workpackages=False,
                                         load_analyse=False)
    if benchmark is None:
        return

    # Change benchmark comment
    if args.append:
        comment = benchmark.comment + args.comment
    else:
        comment = args.comment
    benchmark.comment = re.sub(r"\s+", " ", comment)
    benchmark.write_benchmark_configuration(
        os.path.join(benchmark_folder,
                     jube2.conf.CONFIGURATION_FILENAME), outpath="..")


def gen_parser_conf():
    """Generate dict with parser information"""
    config = (
        (("-V", "--version"),
         {"help": "show version",
          "action": "version",
          "version": "JUBE, version {0}".format(
              jube2.conf.JUBE_VERSION)}),
        (("-v", "--verbose"),
         {"help": "enable verbose console output (use -vv to " +
                  "show stdout during execution and -vvv to " +
                  "show log and stdout)",
          "action": "count",
          "default": 0}),
        (("--debug",),
         {"action": "store_true",
          "help": 'use debugging mode'}),
        (("--force",),
         {"action": "store_true",
          "help": 'skip version check'}),
        (("--strict",),
         {"action": "store_true",
          "help": 'force need for correct version'}),
        (("--devel",),
         {"action": "store_true",
          "help": 'show development related information'})
    )

    return config


def gen_subparser_conf():
    """Generate dict with subparser information"""
    subparser_configuration = dict()

    # run subparser
    subparser_configuration["run"] = {
        "help": "processes benchmark",
        "func": run_new_benchmark,
        "arguments": {
            ("files",):
                {"metavar": "FILE", "nargs": "+", "help": "input file"},
            ("--only-bench",):
                {"nargs": "+", "help": "only run benchmark"},
            ("--not-bench",):
                {"nargs": "+", "help": "do not run benchmark"},
            ("-t", "--tag"):
                {"nargs": "+", "help": "select tags"},
            ("-i", "--id"):
                {"type": int, "help": "use specific benchmark id",
                 "nargs": "+"},
            ("-e", "--error"):
                {"action": "store_true", "help": "exit on error"},
            ("--hide-animation",):
                {"action": "store_true", "help": "hide animations"},
            ("--include-path",):
                {"nargs": "+", "help": "directory containing include files"},
            ("-a", "--analyse"):
                {"action": "store_true", "help": "run analyse"},
            ("-r", "--result"):
                {"action": "store_true", "help": "show results"},
            ("-m", "--comment"):
                {"help": "add comment"},
            ("-o", "--outpath"):
                {"help": "overwrite outpath directory"}
        }
    }

    # continue subparser
    subparser_configuration["continue"] = {
        "help": "continue benchmark",
        "func": continue_benchmarks,
        "arguments": {
            ("dir",):
                {"metavar": "DIRECTORY", "nargs": "?",
                 "help": "benchmark directory", "default": "."},
            ("-i", "--id"):
                {"help": "use benchmarks given by id",
                 "nargs": "+"},
            ("--hide-animation",):
                {"action": "store_true", "help": "hide animations"},
            ("-e", "--error"):
                {"action": "store_true", "help": "exit on error"},
            ("-a", "--analyse"):
                {"action": "store_true", "help": "run analyse"},
            ("-r", "--result"):
                {"action": "store_true", "help": "show results"}
        }
    }

    # analyse subparser
    subparser_configuration["analyse"] = {
        "help": "analyse benchmark",
        "func": analyse_benchmarks,
        "arguments": {
            ("dir",):
                {"metavar": "DIRECTORY", "nargs": "?",
                 "help": "benchmark directory", "default": "."},
            ("-i", "--id"):
                {"help": "use benchmarks given by id",
                 "nargs": "+"},
            ("-u", "--update"):
                {"metavar": "UPDATE_FILE",
                 "help": "update analyse and result configuration"},
            ("--include-path",):
                {"nargs": "+", "help": "directory containing include files"},
            ("-t", "--tag"):
                {"nargs": "+", "help": "select tags"}
        }
    }

    # result subparser
    subparser_configuration["result"] = {
        "help": "show benchmark results",
        "func": benchmarks_results,
        "arguments": {
            ("dir",):
                {"metavar": "DIRECTORY", "nargs": "?",
                 "help": "benchmark directory", "default": "."},
            ("-i", "--id"):
                {"help": "use benchmarks given by id",
                 "nargs": "+"},
            ("-a", "--analyse"):
                {"action": "store_true",
                 "help": "run analyse before creating result"},
            ("-u", "--update"):
                {"metavar": "UPDATE_FILE",
                 "help": "update analyse and result configuration"},
            ("--include-path",):
                {"nargs": "+", "help": "directory containing include files"},
            ("-t", "--tag"):
                {"nargs": '+', "help": "select tags"},
            ("-o", "--only"):
                {"nargs": "+", "metavar": "RESULT_NAME",
                 "help": "only create results given by specific name"},
            ("-r", "--reverse"):
                {"help": "reverse benchmark output order",
                 "action": "store_true"},
            ("-n", "--num"):
                {"type": int, "help": "show only last N benchmarks"},
            ("-s", "--style"):
                {"help": "overwrites table style type",
                 "choices": ["pretty", "csv", "aligned"]}
        }
    }

    # info subparser
    subparser_configuration["info"] = {
        "help": "benchmark information",
        "func": info,
        "arguments": {
            ('dir',):
                {"metavar": "DIRECTORY", "nargs": "?",
                 "help": "benchmark directory", "default": "."},
            ("-i", "--id"):
                {"help": "use benchmarks given by id",
                 "nargs": "+"},
            ("-s", "--step"):
                {"help": "show information for given step", "nargs": "*"},
            ("-p", "--parametrization"):
                {"help": "display only parametrization of given step",
                 "action": "store_true"},
            ("-c", "--csv-parametrization"):
                {"help": "display only parametrization of given step " +
                 "using csv format", "nargs": "?", "default": False,
                 "metavar": "SEPARATOR"}
        }
    }

    # status subparser
    subparser_configuration["status"] = {
        "help": "show benchmark status",
        "func": status,
        "arguments": {
            ('dir',):
                {"metavar": "DIRECTORY", "nargs": "?",
                 "help": "benchmark directory", "default": "."},
            ("-i", "--id"):
                {"help": "use benchmarks given by id",
                 "nargs": "+"}
        }
    }

    # comment subparser
    subparser_configuration["comment"] = {
        "help": "comment handling",
        "func": manipulate_comments,
        "arguments": {
            ('comment',):
                {"help": "comment"},
            ('dir',):
                {"metavar": "DIRECTORY", "nargs": "?",
                 "help": "benchmark directory", "default": "."},
            ("-i", "--id"):
                {"help": "use benchmarks given by id",
                 "nargs": "+"},
            ("-a", "--append"):
                {"help": "append comment to existing one",
                 "action": 'store_true'}
        }
    }

    # remove subparser
    subparser_configuration["remove"] = {
        "help": "remove benchmark or workpackages",
        "func": remove_benchmarks,
        "arguments": {
            ('dir',):
                {"metavar": "DIRECTORY", "nargs": "?",
                 "help": "benchmark directory", "default": "."},
            ("-i", "--id"):
                {"help": "remove benchmarks given by id",
                 "nargs": "+"},
            ("-w", "--workpackage"):
                {"help": "specifc workpackage id to be removed",
                 "nargs": "+"},
            ("-f", "--force"):
                {"help": "force removing, never prompt",
                 "action": "store_true"}
        }
    }

    # update subparser
    subparser_configuration["update"] = {
        "help": "Check if a newer JUBE version is available",
        "func": update_check
    }

    # log subparser
    subparser_configuration["log"] = {
        "help": "show benchmark logs",
        "func": show_log,
        "arguments": {
            ('dir',):
                {"metavar": "DIRECTORY", "nargs": "?",
                 "help": "benchmark directory", "default": "."},
            ('--command', "-c"):
                {"nargs": "+", "help": "show log for this command"},
            ("-i", "--id"):
                {"help": "use benchmarks given by id",
                 "nargs": "+"}
        }
    }

    # completion subparser
    subparser_configuration["complete"] = {
        "help": "generate shell completion "
                'usage: eval "$(jube complete)"',
        "func": complete,
        "arguments": {
            ('--command-name', "-c"):
                {"nargs": 1,
                 "help": "name of command to be completed",
                 "default": [os.path.basename(sys.argv[0])]},
        }
    }

    return subparser_configuration


def _get_args_parser():
    """Create argument parser"""
    parser = argparse.ArgumentParser()

    for args, kwargs in gen_parser_conf():
        parser.add_argument(*args, **kwargs)

    subparsers = parser.add_subparsers(dest="subparser", help='subparsers')

    subparser_configuration = gen_subparser_conf()

    # create subparser out of subparser configuration
    subparser = dict()
    for name, subparser_config in subparser_configuration.items():
        subparser[name] = \
            subparsers.add_parser(
                name, help=subparser_config.get("help", ""),
                description=jube2.help.HELP.get(name, ""),
                formatter_class=argparse.RawDescriptionHelpFormatter)
        subparser[name].set_defaults(func=subparser_config["func"])
        if "arguments" in subparser_config:
            for names, arg in subparser_config["arguments"].items():
                subparser[name].add_argument(*names, **arg)

    # create help key word overview
    help_keys = sorted(list(jube2.help.HELP) + ["ALL"])
    max_word_length = max(map(len, help_keys)) + 4
    # calculate max number of keyword columns
    max_columns = jube2.conf.DEFAULT_WIDTH // max_word_length
    # fill keyword list to match number of columns
    help_keys += [""] * (len(help_keys) % max_columns)
    help_keys = list(zip(*[iter(help_keys)] * max_columns))
    # create overview
    help_overview = jube2.util.output.text_table(help_keys, separator="   ",
                                                 align_right=False)

    # help subparser
    subparser["help"] = \
        subparsers.add_parser(
            'help', help='command help',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="available commands or info elements: \n" +
            help_overview)
    subparser["help"].add_argument('command', nargs='?',
                                   help="command or info element")
    subparser["help"].set_defaults(func=command_help)

    return parser, subparser


def main(command=None):
    """Parse the command line and run the requested command."""

    jube2.help.load_help()
    parser = _get_args_parser()[0]
    if command is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(command)

    jube2.conf.DEBUG_MODE = args.debug
    jube2.conf.VERBOSE_LEVEL = args.verbose

    if jube2.conf.VERBOSE_LEVEL > 0:
        args.hide_animation = True

    # Set new umask if JUBE_GROUP_NAME is used
    current_mask = os.umask(0)
    if (jube2.util.util.check_and_get_group_id() is not None) and \
            (current_mask > 2):
        current_mask = 2
    os.umask(current_mask)

    if args.subparser:
        jube2.log.setup_logging(mode="console",
                                verbose=(jube2.conf.VERBOSE_LEVEL == 1) or
                                        (jube2.conf.VERBOSE_LEVEL == 3))
        if args.devel:
            args.func(args)
        else:
            try:
                args.func(args)
            except Exception as exeption:
                # Catch all possible Exceptions
                LOGGER.error("\n" + str(exeption))
                jube2.log.reset_logging()
                exit(1)
    else:
        parser.print_usage()
    jube2.log.reset_logging()


if __name__ == "__main__":
    main()



# JUBE Benchmarking Environment
# Copyright (C) 2008-2020
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Patternset definition"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import jube2.parameter
import xml.etree.ElementTree as ET

LOGGER = jube2.log.get_logger(__name__)


class Patternset(object):

    """A Patternset stores a set of pattern and derived pattern."""

    def __init__(self, name=""):
        self._name = name
        self._pattern = jube2.parameter.Parameterset("pattern")
        self._derived_pattern = jube2.parameter.Parameterset("derived_pattern")

    def add_pattern(self, pattern):
        """Add a additional pattern to the patternset. Existing pattern using
        the same name will be overwritten"""
        if pattern.derived:
            if pattern in self._pattern:
                self._pattern.delete_parameter(pattern)
            self._derived_pattern.add_parameter(pattern)
        else:
            if pattern in self._derived_pattern:
                self._derived_pattern.delete_parameter(pattern)
            self._pattern.add_parameter(pattern)

    @property
    def pattern_storage(self):
        """Return the pattern storage"""
        return self._pattern

    @property
    def derived_pattern_storage(self):
        """Return the derived pattern storage"""
        return self._derived_pattern

    def etree_repr(self):
        """Return etree object representation"""
        patternset_etree = ET.Element('patternset')
        patternset_etree.attrib["name"] = self._name
        for pattern in self._pattern:
            patternset_etree.append(
                pattern.etree_repr())
        for pattern in self._derived_pattern:
            patternset_etree.append(
                pattern.etree_repr())
        return patternset_etree

    def add_patternset(self, patternset):
        """Add all pattern from given patternset to the current one"""
        self._pattern.add_parameterset(patternset.pattern_storage)
        self._derived_pattern.add_parameterset(
            patternset.derived_pattern_storage)

    def pattern_substitution(self, parametersets=None):
        """Run pattern substitution using additional parameterset"""
        if parametersets is None:
            parametersets = list()
        self._pattern.parameter_substitution(
            additional_parametersets=parametersets,
            final_sub=True)

    def derived_pattern_substitution(self, parametersets=None):
        """Run derived pattern substitution using additional parameterset"""
        if parametersets is None:
            parametersets = list()
        self._derived_pattern.parameter_substitution(
            additional_parametersets=parametersets,
            final_sub=True)

    @property
    def name(self):
        """Get patternset name"""
        return self._name

    def copy(self):
        """Returns a copy of the Parameterset"""
        new_patternset = Patternset(self._name)
        new_patternset.add_patternset(self)
        return new_patternset

    def is_compatible(self, patternset):
        """Two Patternsets are compatible, if all pattern storages are
        compatible"""
        return self.pattern_storage.is_compatible(
            patternset.pattern_storage) and \
            self.pattern_storage.is_compatible(
                patternset.derived_pattern_storage) and \
            self.derived_pattern_storage.is_compatible(
                patternset.derived_pattern_storage) and \
            self.derived_pattern_storage.is_compatible(
                patternset.pattern_storage)

    def get_incompatible_pattern(self, patternset):
        """Return a set of incompatible pattern names between the current
        and the given parameterset"""
        result = set()
        result.update(self.pattern_storage.get_incompatible_parameter(
            patternset.pattern_storage))
        result.update(self.pattern_storage.get_incompatible_parameter(
            patternset.derived_pattern_storage))
        result.update(self.derived_pattern_storage.get_incompatible_parameter(
            patternset.pattern_storage))
        result.update(self.derived_pattern_storage.get_incompatible_parameter(
            patternset.derived_pattern_storage))

        return result

    def __repr__(self):
        return "Patternset: pattern:{0} derived pattern:{1}".format(
            dict([[pattern.name, pattern.value]
                  for pattern in self._pattern]),
            dict([[pattern.name, pattern.value]
                  for pattern in self._derived_pattern]))

    def __contains__(self, pattern):
        if isinstance(pattern, Pattern):
            if pattern.name in self._pattern:
                return pattern.is_equivalent(
                    self._pattern[pattern.name])
            elif pattern.name in self._derived_pattern:
                return pattern.is_equivalent(
                    self._derived_pattern[pattern.name])
            else:
                return False
        else:
            return (pattern in self._pattern) or \
                   (pattern in self._derived_pattern)

    def __getitem__(self, name):
        """Returns pattern given by name. Is pattern not found, None will
        be returned"""
        if name in self._pattern:
            return self._pattern[name]
        elif name in self._derived_pattern:
            return self._derived_pattern[name]
        else:
            return None


class Pattern(jube2.parameter.StaticParameter):

    """A pattern can be used to scan a result file, using regular expression,
    or to represent a derived pattern."""

    def __init__(self, name, value, pattern_mode="pattern",
                 content_type="string", unit="", default=None, dotall=False):
        self._derived = pattern_mode != "pattern"

        if not self._derived:
            pattern_mode = "text"

        self._default = default

        self._dotall = dotall

        # Unicode conversion
        value = "" + value

        jube2.parameter.StaticParameter.__init__(
            self, name, value, parameter_type=content_type,
            parameter_mode=pattern_mode)

        self._unit = unit

    @property
    def derived(self):
        """pattern is a derived pattern"""
        return self._derived

    @property
    def content_type(self):
        """Return pattern type"""
        return self._type

    @property
    def default_value(self):
        """Return pattern default value"""
        return self._default

    @property
    def dotall(self):
        """Return pattern dot regex handling"""
        return self._dotall

    @property
    def unit(self):
        """Return unit"""
        return self._unit

    def substitute_and_evaluate(self, parametersets=None,
                                final_sub=False, no_templates=True,
                                force_evaluation=False):
        """Substitute all variables inside the pattern value by using the
        parameter inside the given parameterset and additional_parameterset.
        final_sub marks the last substitution.

        Return the new pattern and a boolean value which represent a change
        of value
        """
        try:
            # To take care of default values for derived pattern sets, always
            # run final_sub instead of force_evaluation. Otherwise no error
            # will be thrown. Only using the final_sub setup is too late
            # because the default pattern might be used within another derived
            # pattern
            if (self._mode in jube2.conf.ALLOWED_SCRIPTTYPES and
                    force_evaluation and self._default is not None):
                final_sub = True
                force_evaluation = False
            param, changed = \
                jube2.parameter.StaticParameter.substitute_and_evaluate(
                    self, parametersets, final_sub, no_templates,
                    force_evaluation)
        except RuntimeError as re:
            LOGGER.debug(str(re).replace("parameter", "pattern"))
            if self._default is not None:
                value = self._default
            elif self._type in ["int", "float"]:
                value = "nan"
            else:
                value = ""
            pattern = Pattern(
                self._name, value, "text", self._type, self._unit,
                dotall=self._dotall)
            pattern.based_on = self
            return pattern, True

        if changed:
            # Convert parameter to pattern
            if not self.derived:
                pattern_mode = "pattern"
            else:
                pattern_mode = param.mode
            pattern = Pattern(param.name, param.value, pattern_mode,
                              param.parameter_type, self._unit,
                              dotall=self._dotall)
            pattern.based_on = param.based_on
        else:
            pattern = param
        return pattern, changed

    def etree_repr(self, use_current_selection=False):
        """Return etree object representation"""
        pattern_etree = ET.Element('pattern')
        pattern_etree.attrib["name"] = self._name
        pattern_etree.attrib["type"] = self._type
        pattern_etree.attrib["dotall"] = str(self._dotall)
        if self._default is not None:
            pattern_etree.attrib["default"] = self._default
        if not self._derived:
            pattern_etree.attrib["mode"] = "pattern"
        else:
            pattern_etree.attrib["mode"] = self._mode

        if self._unit != "":
            pattern_etree.attrib["unit"] = self._unit
        pattern_etree.text = self.value
        return pattern_etree

    def __repr__(self):
        return "Pattern({0})".format(self.__dict__)


def get_jube_pattern():
    """Return jube internal patternset"""
    patternset = Patternset()
    # Pattern for integer number
    patternset.add_pattern(Pattern("jube_pat_int", r"([+-]?\d+)"))
    # Pattern for integer number, no ()
    patternset.add_pattern(Pattern("jube_pat_nint", r"(?:[+-]?\d+)"))
    # Pattern for floating point number
    patternset.add_pattern(
        Pattern("jube_pat_fp", r"([+-]?(?:\d*\.?\d+(?:[eE][-+]?\d+)?|\d+\.))"))
    # Pattern for floating point number, no ()
    patternset.add_pattern(
        Pattern("jube_pat_nfp",
                r"(?:[+-]?(?:\d*\.?\d+(?:[eE][-+]?\d+)?|\d+\.))"))
    # Pattern for word (all noblank characters)
    patternset.add_pattern(Pattern("jube_pat_wrd", r"(\S+)"))
    # Pattern for word (all noblank characters), no ()
    patternset.add_pattern(Pattern("jube_pat_nwrd", r"(?:\S+)"))
    # Pattern for blank space (variable length)
    patternset.add_pattern(Pattern("jube_pat_bl", r"(?:\s+)"))
    return patternset



# JUBE Benchmarking Environment
# Copyright (C) 2008-2020
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""The Benchmark class manages the benchmark process"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import xml.etree.ElementTree as ET
import xml.dom.minidom as DOM
import os
import stat
import pprint
import shutil
import itertools
import jube2.parameter
import jube2.util.util
import jube2.util.output
import jube2.conf
import jube2.log

LOGGER = jube2.log.get_logger(__name__)


class Benchmark(object):

    """The Benchmark class contains all data to run a benchmark"""

    def __init__(self, name, outpath, parametersets, substitutesets,
                 filesets, patternsets, steps, analyser, results,
                 results_order, comment="", tags=None, file_path_ref="."):
        self._name = name
        self._outpath = outpath
        self._parametersets = parametersets
        self._substitutesets = substitutesets
        self._filesets = filesets
        self._patternsets = patternsets
        self._steps = steps
        self._analyser = analyser
        for analyser in self._analyser.values():
            analyser.benchmark = self
        self._results = results
        self._results_order = results_order
        for result in self._results.values():
            result.benchmark = self
        self._workpackages = dict()
        self._work_stat = jube2.util.util.WorkStat()
        self._comment = comment
        self._id = -1
        self._file_path_ref = file_path_ref
        if tags is None:
            self._tags = set()
        else:
            self._tags = tags

    @property
    def name(self):
        """Return benchmark name"""
        return self._name

    @property
    def comment(self):
        """Return comment string"""
        return self._comment

    @property
    def tags(self):
        """Return set of tags"""
        return self._tags

    @comment.setter
    def comment(self, new_comment):
        """Set new comment string"""
        self._comment = new_comment

    @property
    def parametersets(self):
        """Return parametersets"""
        return self._parametersets

    @property
    def patternsets(self):
        """Return patternsets"""
        return self._patternsets

    @property
    def analyser(self):
        """Return analyser"""
        return self._analyser

    @property
    def results(self):
        """Return results"""
        return self._results

    @property
    def results_order(self):
        """Return results_order"""
        return self._results_order

    @property
    def file_path_ref(self):
        """Get file path reference"""
        return self._file_path_ref

    @file_path_ref.setter
    def file_path_ref(self, file_path_ref):
        """Set file path reference"""
        self._file_path_ref = file_path_ref

    @property
    def outpath(self):
        """Return outpath"""
        return self._outpath

    @outpath.setter
    def outpath(self, new_outpath):
        """Overwrite outpath"""
        self._outpath = new_outpath

    @property
    def substitutesets(self):
        """Return substitutesets"""
        return self._substitutesets

    @property
    def workpackages(self):
        """Return workpackages"""
        return self._workpackages

    def add_tags(self, other_tags):
        if other_tags is not None:
            self._tags = self._tags.union(set(other_tags))

    def workpackage_by_id(self, wp_id):
        """Search and return a benchmark workpackage by its wp_id"""
        for stepname in self._workpackages:
            for workpackage in self._workpackages[stepname]:
                if workpackage.id == wp_id:
                    return workpackage
        return None

    def remove_workpackage(self, workpackage_to_delete):
        """Remove a specifc workpackage"""
        stepname = workpackage_to_delete.step.name
        if stepname in self._workpackages and \
                workpackage_to_delete in self._workpackages[stepname]:
            self._workpackages[stepname].remove(workpackage_to_delete)

    @property
    def work_stat(self):
        """Return work queue"""
        return self._work_stat

    @property
    def filesets(self):
        """Return filesets"""
        return self._filesets

    def delete_bench_dir(self):
        """Delete all data inside benchmark directory"""
        if os.path.exists(self.bench_dir):
            shutil.rmtree(self.bench_dir, ignore_errors=True)

    @property
    def steps(self):
        """Return steps"""
        return self._steps

    @property
    def workpackage_status(self):
        """Retun workpackage information dict"""
        result_dict = dict()
        for stepname in self._workpackages:
            result_dict[stepname] = {"all": 0,
                                     "open": 0,
                                     "wait": 0,
                                     "error": 0,
                                     "done": 0}
            for workpackage in self._workpackages[stepname]:
                result_dict[stepname]["all"] += 1
                if workpackage.done:
                    result_dict[stepname]["done"] += 1
                elif workpackage.error:
                    result_dict[stepname]["error"] += 1
                elif workpackage.started:
                    result_dict[stepname]["wait"] += 1
                else:
                    result_dict[stepname]["open"] += 1
        return result_dict

    @property
    def benchmark_status(self):
        """Retun global workpackage information dict"""
        result_dict = {"all": 0,
                       "open": 0,
                       "wait": 0,
                       "error": 0,
                       "done": 0}

        for status in self.workpackage_status.values():
            result_dict["all"] += status["all"]
            result_dict["open"] += status["open"]
            result_dict["wait"] += status["wait"]
            result_dict["error"] += status["error"]
            result_dict["done"] += status["done"]

        return result_dict

    @property
    def id(self):
        """Return benchmark id"""
        return self._id

    @id.setter
    def id(self, new_id):
        """Set new benchmark id"""
        self._id = new_id

    def get_jube_parameterset(self):
        """Return parameterset which contains benchmark related
        information"""
        parameterset = jube2.parameter.Parameterset()
        # benchmark id
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter(
                "jube_benchmark_id", str(self._id), parameter_type="int",
                update_mode=jube2.parameter.JUBE_MODE))

        # benchmark id with padding
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_benchmark_padid",
                             jube2.util.util.id_dir("", self._id),
                             parameter_type="string",
                             update_mode=jube2.parameter.JUBE_MODE))

        # benchmark name
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_benchmark_name", self._name,
                             update_mode=jube2.parameter.JUBE_MODE))

        # benchmark home
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_benchmark_home",
                             os.path.abspath(self._file_path_ref),
                             update_mode=jube2.parameter.JUBE_MODE))

        # benchmark rundir
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_benchmark_rundir",
                             os.path.abspath(self.bench_dir),
                             update_mode=jube2.parameter.JUBE_MODE))

        timestamps = jube2.util.util.read_timestamps(
            os.path.join(self.bench_dir, jube2.conf.TIMESTAMPS_INFO))

        # benchmark start
        parameterset.add_parameter(
            jube2.parameter.Parameter.create_parameter(
                "jube_benchmark_start",
                timestamps.get("start", "").replace(" ", "T"),
                update_mode=jube2.parameter.JUBE_MODE))

        return parameterset

    def etree_repr(self, new_cwd=None):
        """Return etree object representation"""
        benchmark_etree = ET.Element("benchmark")
        if len(self._comment) > 0:
            comment_element = ET.SubElement(benchmark_etree, "comment")
            comment_element.text = self._comment
        benchmark_etree.attrib["name"] = self._name
        # Modify file_path_ref and outpath to be relativly correct towards
        # new configuration file position
        if new_cwd is not None:
            benchmark_etree.attrib["file_path_ref"] = \
                os.path.relpath(self._file_path_ref, new_cwd)
            if not os.path.isabs(self._outpath):
                benchmark_etree.attrib["outpath"] = \
                    os.path.relpath(self._outpath, new_cwd)
            else:
                benchmark_etree.attrib["outpath"] = self._outpath

        for parameterset in self._parametersets.values():
            benchmark_etree.append(parameterset.etree_repr())
        for substituteset in self._substitutesets.values():
            benchmark_etree.append(substituteset.etree_repr())
        for fileset in self._filesets.values():
            benchmark_etree.append(fileset.etree_repr())
        for patternset in self._patternsets.values():
            benchmark_etree.append(patternset.etree_repr())
        for step in self._steps.values():
            benchmark_etree.append(step.etree_repr())
        for analyser in self._analyser.values():
            benchmark_etree.append(analyser.etree_repr())
        for result_name in self._results_order:
            result = self._results[result_name]
            benchmark_etree.append(result.etree_repr())
        return benchmark_etree

    def __repr__(self):
        return pprint.pformat(self.__dict__)

    def _create_initial_workpackages(self):
        """Create initial workpackages of current benchmark and create graph
        structure."""
        self._workpackages = dict()
        self._work_stat = jube2.util.util.WorkStat()

        # Create workpackage storage
        for step_name in self._steps:
            self._workpackages[step_name] = list()

        # Create initial workpackages
        for step in self._steps.values():
            if len(step.depend) == 0:
                new_workpackages = \
                    self._create_new_workpackages_with_parents(step)
                self._workpackages[step.name] += new_workpackages
                for workpackage in new_workpackages:
                    workpackage.queued = True
                    self._work_stat.put(workpackage)

    def analyse(self, show_info=True, specific_analyser_name=None):
        """Run analyser"""

        if show_info:
            LOGGER.info(">>> Start analyse")

        if specific_analyser_name is not None and \
                specific_analyser_name in self._analyser:
            self._analyser[specific_analyser_name].analyse()
        else:
            for analyser in self._analyser.values():
                analyser.analyse()
        if ((not jube2.conf.DEBUG_MODE) and
                (os.access(self.bench_dir, os.W_OK))):
            self.write_analyse_data(os.path.join(self.bench_dir,
                                                 jube2.conf.ANALYSE_FILENAME))
        if show_info:
            LOGGER.info(">>> Analyse finished")

    def create_result(self, only=None, show=False, data_list=None, style=None):
        """Show benchmark result"""
        if only is None:
            only = [result_name for result_name in self._results]
        if data_list is None:
            data_list = list()
        for result_name in self._results_order:
            result = self._results[result_name]
            if result.name in only:
                result_data = result.create_result_data(style)
                if result.result_dir is None:
                    result_dir = os.path.join(self.bench_dir,
                                              jube2.conf.RESULT_DIRNAME)
                else:
                    result_dir = result.result_dir
                    result_dir = os.path.expanduser(result_dir)
                    result_dir = os.path.expandvars(result_dir)
                    result_dir = jube2.util.util.id_dir(
                        os.path.join(self.file_path_ref, result_dir), self.id)
                if (not os.path.exists(result_dir)) and \
                   (not jube2.conf.DEBUG_MODE):
                    try:
                        os.makedirs(result_dir)
                    except OSError:
                        pass
                if ((not jube2.conf.DEBUG_MODE) and
                        (os.path.exists(result_dir)) and
                        (os.access(result_dir, os.W_OK))):
                    filename = os.path.join(result_dir,
                                            "{0}.dat".format(result.name))
                else:
                    filename = None
                result_data.create_result(show=show, filename=filename)

                if result_data in data_list:
                    data_list[data_list.index(result_data)].add_result_data(
                        result_data)
                else:
                    data_list.append(result_data)
        return data_list

    def update_analyse_and_result(self, new_patternsets, new_analyser,
                                  new_results, new_results_order, new_cwd):
        """Update analyser and result data"""
        if os.path.exists(self.bench_dir):
            LOGGER.debug("Update analyse and result data")
            self._patternsets = new_patternsets
            old_analyser = self._analyser
            self._analyser = new_analyser
            self._results = new_results
            self._results_order = new_results_order
            for analyser in self._analyser.values():
                if analyser.name in old_analyser:
                    analyser.analyse_result = \
                        old_analyser[analyser.name].analyse_result
                analyser.benchmark = self
            for result in self._results.values():
                result.benchmark = self
                # change result dir position relative to cwd
                if (result.result_dir is not None) and \
                   (new_cwd is not None) and \
                   (not os.path.isabs(result.result_dir)):
                    result.result_dir = \
                        os.path.join(new_cwd, result.result_dir)
            if ((not jube2.conf.DEBUG_MODE) and
                    (os.access(self.bench_dir, os.W_OK))):
                self.write_benchmark_configuration(
                    os.path.join(self.bench_dir,
                                 jube2.conf.CONFIGURATION_FILENAME),
                    outpath="..")

    def write_analyse_data(self, filename):
        """All analyse data will be written to given file
        using xml representation"""
        # Create root-tag and append analyser
        analyse_etree = ET.Element("analyse")
        for analyser_name in self._analyser:
            analyser_etree = ET.SubElement(analyse_etree, "analyser")
            analyser_etree.attrib["name"] = analyser_name
            for etree in self._analyser[analyser_name].analyse_etree_repr():
                analyser_etree.append(etree)
        xml = jube2.util.output.element_tree_tostring(
            analyse_etree, encoding="UTF-8")
        # Using dom for pretty-print
        dom = DOM.parseString(xml.encode("UTF-8"))
        fout = open(filename, "wb")
        fout.write(dom.toprettyxml(indent="  ", encoding="UTF-8"))
        fout.close()

    def _create_new_workpackages_for_workpackage(self, workpackage):
        """Create and return new workpackages if given workpackage
        was finished."""
        all_new_workpackages = list()
        if not workpackage.done or len(workpackage.children) > 0:
            return all_new_workpackages
        LOGGER.debug(("Create new workpackages for workpackage"
                      " {0}({1})").format(
            workpackage.step.name, workpackage.id))
        # Search for dependent steps
        dependent_steps = [step for step in self._steps.values() if
                           workpackage.step.name in step.depend]

        # Search for possible workpackage parents
        for dependent_step in dependent_steps:
            parent_workpackages = [[
                parent_workpackage for parent_workpackage in
                self._workpackages[step_name] if parent_workpackage.done]
                for step_name in dependent_step.depend
                if (step_name in self._workpackages) and
                   (step_name != workpackage.step.name)]
            parent_workpackages.append([workpackage])

            # Create all possible parent combinations
            workpackage_combinations = \
                [iterator for iterator in
                 itertools.product(*parent_workpackages)]
            possible_combination = len(workpackage_combinations)
            for workpackage_combination in workpackage_combinations:
                new_workpackages = self._create_new_workpackages_with_parents(
                    dependent_step, workpackage_combination)
                if len(new_workpackages) > 0:
                    possible_combination -= 1

                # Create links: parent workpackages -> new children
                for new_workpackage in new_workpackages:
                    for parent in workpackage_combination:
                        parent.add_children(new_workpackage)

                self._workpackages[dependent_step.name] += new_workpackages
                all_new_workpackages += new_workpackages
            if possible_combination > 0:
                LOGGER.debug(("  {0} workpackages combinations were skipped"
                              " while checking possible parent combinations"
                              " for step {1}").format(possible_combination,
                                                      dependent_step.name))

        LOGGER.debug("  {0} new workpackages created".format(
            len(all_new_workpackages)))
        return all_new_workpackages

    def _create_new_workpackages_with_parents(self, step,
                                              parent_workpackages=None):
        """Create workpackages with given parent combination"""
        if parent_workpackages is None:
            parent_workpackages = list()
        # Combine and check parent parametersets
        parameterset = jube2.parameter.Parameterset()
        incompatible_parameter_names = set()
        for parent_workpackage in parent_workpackages:
            # Check weather parameter combination is possible or not.
            # JUBE Parameter can be ignored
            incompatible_parameter_names = incompatible_parameter_names.union(
                parameterset.get_incompatible_parameter(
                    parent_workpackage.parameterset,
                    update_mode=jube2.parameter.JUBE_MODE))
            parameterset.add_parameterset(
                parent_workpackage.parameterset)

        # Sort parent workpackges after total iteration number and name
        sorted_parents = list(parent_workpackages)
        sorted_parents.sort(key=lambda x: x.step.name)
        sorted_parents.sort(key=lambda x: x.step.iterations)

        iteration_base = 0
        for i, parent in enumerate(sorted_parents):
            if i == 0:
                iteration_base = parent.iteration
            else:
                iteration_base = \
                    parent.step.iterations * iteration_base + parent.iteration

        parameterset.remove_jube_parameter()

        # Create new workpackages
        new_workpackages = step.create_workpackages(
            self, parameterset,
            iteration_base=iteration_base,
            parents=parent_workpackages,
            incompatible_parameters=incompatible_parameter_names)

        # Update iteration sibling connections
        if len(parent_workpackages) > 0 and len(new_workpackages) > 0:
            for sibling in parent_workpackages[0].iteration_siblings:
                if sibling != parent_workpackages[0]:
                    for child in sibling.children:
                        for workpackage in new_workpackages:
                            if workpackage.parameterset.is_compatible(
                                    child.parameterset,
                                    update_mode=jube2.parameter.JUBE_MODE):
                                workpackage.iteration_siblings.add(child)
                                child.iteration_siblings.add(workpackage)

        return new_workpackages

    def new_run(self):
        """Create workpackage structure and run benchmark"""
        # Check benchmark consistency
        LOGGER.debug("Start consistency check")
        jube2.util.util.consistency_check(self)

        # Create benchmark directory
        LOGGER.debug("Create benchmark directory")
        self._create_bench_dir()

        # Change logfile
        jube2.log.change_logfile_name(os.path.join(
            self.bench_dir, jube2.conf.LOGFILE_RUN_NAME))
        # Move parse logfile into benchmark folder
        if os.path.isfile(os.path.join(self._file_path_ref,
                                       jube2.conf.DEFAULT_LOGFILE_NAME)):
            os.rename(os.path.join(self._file_path_ref,
                                   jube2.conf.DEFAULT_LOGFILE_NAME),
                      os.path.join(self.bench_dir,
                                   jube2.conf.LOGFILE_PARSE_NAME))

        # Reset Workpackage counter
        jube2.workpackage.Workpackage.id_counter = 0

        # Create initial workpackages
        LOGGER.debug("Create initial workpackages")
        self._create_initial_workpackages()

        # Store workpackage information
        LOGGER.debug("Store initial workpackage information")
        self.write_workpackage_information(
            os.path.join(self.bench_dir, jube2.conf.WORKPACKAGES_FILENAME))

        LOGGER.debug("Start benchmark run")
        self.run()

    def run(self):
        """Run benchmark"""
        title = "benchmark: {0}".format(self._name)
        title += "\nid: {0}".format(self._id)
        if jube2.conf.DEBUG_MODE:
            title += " ---DEBUG_MODE---"
        title += "\n\n{0}".format(self._comment)
        infostr = jube2.util.output.text_boxed(title)
        LOGGER.info(infostr)

        if not jube2.conf.HIDE_ANIMATIONS:
            print("\nRunning workpackages (#=done, 0=wait, E=error):")
            status = self.benchmark_status
            jube2.util.output.print_loading_bar(
                status["done"], status["all"], status["wait"], status["error"])

        # Handle all workpackages in given order
        while not self._work_stat.empty():
            workpackage = self._work_stat.get()
            if not workpackage.done:
                workpackage.run()
            self._create_new_workpackages_for_workpackage(workpackage)

            # Update queues (move waiting workpackages to work queue
            # if possible)
            self._work_stat.update_queues(workpackage)

            if not jube2.conf.HIDE_ANIMATIONS:
                status = self.benchmark_status
                jube2.util.output.print_loading_bar(
                    status["done"], status["all"], status["wait"],
                    status["error"])
            workpackage.queued = False

            for mode in ("only_started", "all"):
                for child in workpackage.children:
                    all_done = True
                    for parent in child.parents:
                        all_done = all_done and parent.done
                    if all_done:
                        if (mode == "only_started" and child.started) or \
                           (mode == "all" and (not child.queued)):
                            child.queued = True
                            self._work_stat.put(child)
        # Store workpackage information
        self.write_workpackage_information(
            os.path.join(self.bench_dir, jube2.conf.WORKPACKAGES_FILENAME))

        print("\n")

        status_data = [("stepname", "all", "open", "wait", "error", "done")]
        status_data += [(stepname, str(_status["all"]), str(_status["open"]),
                         str(_status["wait"]), str(_status["error"]),
                         str(_status["done"]))
                        for stepname, _status in
                        self.workpackage_status.items()]
        LOGGER.info(jube2.util.output.text_table(
            status_data, use_header_line=True, indent=2))

        LOGGER.info("\n>>>> Benchmark information and " +
                    "further useful commands:")
        LOGGER.info(">>>>       id: {0}".format(self._id))
        LOGGER.info(">>>>   handle: {0}".format(self._outpath))
        LOGGER.info(">>>>      dir: {0}".format(self.bench_dir))

        status = self.benchmark_status
        if status["all"] != status["done"]:
            LOGGER.info((">>>> continue: jube continue {0} " +
                         "--id {1}").format(self._outpath, self._id))
        LOGGER.info((">>>>  analyse: jube analyse {0} " +
                     "--id {1}").format(self._outpath, self._id))
        LOGGER.info((">>>>   result: jube result {0} " +
                     "--id {1}").format(self._outpath, self._id))
        LOGGER.info((">>>>     info: jube info {0} " +
                     "--id {1}").format(self._outpath, self._id))
        LOGGER.info((">>>>      log: jube log {0} " +
                     "--id {1}").format(self._outpath, self._id))
        LOGGER.info(jube2.util.output.text_line() + "\n")

    def _create_bench_dir(self):
        """Create the directory for a benchmark."""
        # Get group_id if available (given by JUBE_GROUP_NAME)
        group_id = jube2.util.util.check_and_get_group_id()
        # Check if outpath exists
        if not (os.path.exists(self._outpath) and
                os.path.isdir(self._outpath)):
            os.makedirs(self._outpath)
            if group_id is not None:
                os.chown(self._outpath, os.getuid(), group_id)
        # Generate unique ID in outpath
        if self._id < 0:
            self._id = jube2.util.util.get_current_id(self._outpath) + 1
        if os.path.exists(self.bench_dir):
            raise RuntimeError("Benchmark directory \"{0}\" already exists"
                               .format(self.bench_dir))

        os.makedirs(self.bench_dir)
        # If JUBE_GROUP_NAME is given, set GID-Bit and change group
        if group_id is not None:
            os.chown(self.bench_dir, os.getuid(), group_id)
            os.chmod(self.bench_dir,
                     os.stat(self.bench_dir).st_mode | stat.S_ISGID)
        self.write_benchmark_configuration(
            os.path.join(self.bench_dir, jube2.conf.CONFIGURATION_FILENAME),
            outpath="..")
        jube2.util.util.update_timestamps(os.path.join(
            self.bench_dir, jube2.conf.TIMESTAMPS_INFO), "start", "change")

    def write_benchmark_configuration(self, filename, outpath=None):
        """The current benchmark configuration will be written to given file
        using xml representation"""
        # Create root-tag and append single benchmark
        benchmarks_etree = ET.Element("jube")
        benchmarks_etree.attrib["version"] = jube2.conf.JUBE_VERSION
        # Store tag information
        if len(self._tags) > 0:
            selection_etree = ET.SubElement(benchmarks_etree, "selection")
            for tag in self._tags:
                tag_etree = ET.SubElement(selection_etree, "tag")
                tag_etree.text = tag

        benchmark_etree = self.etree_repr(new_cwd=self.bench_dir)
        if outpath is not None:
            benchmark_etree.attrib["outpath"] = outpath

        benchmarks_etree.append(benchmark_etree)
        xml = jube2.util.output.element_tree_tostring(
            benchmarks_etree, encoding="UTF-8")
        # Using dom for pretty-print
        dom = DOM.parseString(xml.encode('UTF-8'))
        fout = open(filename, "wb")
        fout.write(dom.toprettyxml(indent="  ", encoding="UTF-8"))
        fout.close()

    def reset_all_workpackages(self):
        """Reset workpackage state"""
        for workpackages in self._workpackages.values():
            for workpackage in workpackages:
                workpackage.done = False

    def write_workpackage_information(self, filename):
        """All workpackage information will be written to given file
        using xml representation"""
        # Create root-tag and append workpackages
        workpackages_etree = ET.Element("workpackages")
        for workpackages in self._workpackages.values():
            for workpackage in workpackages:
                workpackages_etree.append(workpackage.etree_repr())
        xml = jube2.util.output.element_tree_tostring(
            workpackages_etree, encoding="UTF-8")
        # Using dom for pretty-print
        dom = DOM.parseString(xml.encode("UTF-8"))
        fout = open(filename, "wb")
        fout.write(dom.toprettyxml(indent="  ", encoding="UTF-8"))
        fout.close()

    def set_workpackage_information(self, workpackages, work_stat):
        """Set new workpackage information"""
        self._workpackages = workpackages
        self._work_stat = work_stat

    @property
    def bench_dir(self):
        """Return benchmark directory"""
        return jube2.util.util.id_dir(self._outpath, self._id)
